#ifndef BrashRedirection_header_included_p
#define BrashRedirection_header_included_p
//
// Copyright 2015-2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include "BrashException.h"

#include <cxxtls/stringexpression.h>
#include <cxxtls/ScriptLineStream.h>
#include <cxxtls/scriptthread.h>


#include <utility> // for std::pair

struct RedirectionException {};   // non-terminating exceptions that really are just syntax errors
                              
DECLARE_CXXTLS_PARENTED_EXCEPTION(InvalidRedirection, RedirectionException);
DECLARE_CXXTLS_PARENTED_EXCEPTION(NonAllowedRedirectionException, RedirectionException);      // cannot swap input and output
DECLARE_CXXTLS_PARENTED_EXCEPTION(EmptyRedirectionFilenameException, RedirectionException);   // some operations must have a valid filename
DECLARE_CXXTLS_PARENTED_EXCEPTION(CannotDupBadFileException, RedirectionException);           // attempt to duplicate a bad file handle
DECLARE_CXXTLS_PARENTED_EXCEPTION(CannotOpenTargetFileException, RedirectionException);       // just can't open the desired file

struct RedirType
    ///  A representation of an io redirection type.  Not the names of the files,
    ///  but the kind of redirection.  Input, Output, and/or duplication of some
    ///  other file handle.
{
    bool isInput_;     // input vs output redirection
    bool isAppend_;    // << or >>
    int fileNo_;       // the file handle being redirected. Typically, 0, 1, or 2, but others are possible
    int otherFile_;    // redirecting as a copy of some other file hande

    // if otherFile is -1, then no redirection is occuring.

    RedirType(int fileno, bool input, bool append, int otherFile)
    : isInput_(input)
    , isAppend_(append)
    , fileNo_(fileno)
    , otherFile_(otherFile)
    {
    }

    typedef std::list< std::pair<RedirType, cxxtls::StringExpression*> > REDIRECTION_LIST;

    bool operator< (RedirType const &rhs) const  // probably not useful.  redirections have a manually defined order
                                                 // and no comparison operator can overcome that.
    {
       return fileNo_ < rhs.fileNo_;
    }

    RedirType(std::string const &evaluatedExpression);
        ///<  Construct a redirection definition from a string that is identified
        ///   by StringExpr::parse() with the attribute "isRedir()".  
        ///
        ///   To use this constructor:
        ///
        ///   @li call StringExpr::parse()
        ///   @li check the isRedir() flag
        ///   @li if it is true, then evaluate the expression in the current context
        ///       and pass the result to this constructor.

    static bool isNumberedRedirection(std::string::const_iterator &f, std::string::const_iterator const &l, int &fileno);


    static std::pair<std::string, REDIRECTION_LIST*> 
        parseOnlyRedirections(std::string::const_iterator &f,
                              std::string::const_iterator const &l
                             );
    ///<  Parse a trailing sequence of redirection tokens from an input stream up to the end of the
    ///   statement.  This function is meant to be used after the trailing token in
    ///   each of the following statement types:
    ///   @li if ... then ... else ... fi <trailing redirection tokens>
    ///   @li where ... do ... done <trailing redirection tokens>
    ///   @li for ... do ... done <trailing redirection tokens>
    ///   @li { ... } <trailing redirection tokens>
    ///   @li ( ... ) <trailing redirection tokens>
    ///
    ///   @return pair<string, REDIRECTION_LIST*>
    ///
    ///   If the redirection list is a null pointer, then string contains an error message.  Otherwise
    ///   not.  The caller is responsible fo deleting the redirection list.

    static void redirectFileHandles(cxxtls::ScriptThread::ThreadContext *workingContext_, RedirType::REDIRECTION_LIST const *redirs );

};

#endif
