#ifndef FILE_HEADER_INCLUDED
#define FILE_HEADER_INCLUDED

//
// Copyright 2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <string>
#include <string.h>
#include <memory>
#include <sys/types.h>
#include <time.h>
#include <sys/stat.h>
#include <iostream>
#include <list>
#include <set>
#include <memory>
#include <cxxtls/strtool.h>

namespace cxxtls
{

///@file
///
/// A collection of classes and functions that simplify file system
/// access and the use of data objects, such as file names, operating
/// system status information, etc.

//========================================================================================
//========================================================================================
//========================================================================================

struct FileTime
  ///
  /// A portable representation of a file's (local) time
  ///
  /// A FileTime object hold's a time information appropos to a
  /// file.  It has member functions typically used in printing and
  /// displaying the time associated with files.
  ///
{
  int year;     ///<  year minus 1900
  int month;    ///<  month, 0 is january
  int day;      ///<  day of the month, starting with 1
  int hour;     ///<  hour, 0 - 23
  int minute;   ///<  minute, 0 - 59
  int second;   ///<  second, 0 - 59

  friend bool operator<(FileTime const &l, FileTime const &r)
  {
      if(l.year < r.year)
         return true;
      else
      if(l.year > r.year)
        return false;


      if(l.month < r.month)
         return true;
      else
      if(l.month > r.month)
        return false;


      if(l.day < r.day)
         return true;
      else
      if(l.day > r.day)
        return false;


      if(l.hour < r.hour)
         return true;
      else
      if(l.hour > r.hour)
        return false;


      if(l.minute < r.minute)
         return true;
      else
      if(l.minute > r.minute)
        return false;


      if(l.second < r.second)
         return true;
      else
      if(l.second > r.year)
        return false;


      return false;
  }


  operator std::string() const;
    /// convert the time to a string (long format )

  friend std::ostream& operator<< (std::ostream& o, FileTime const &r)
    /// prints the time to an output stream (long format)
  {
    return o << std::string(r);
  }

  FileTime()       ///  FileTime()  default constructor
  : year(0),
    month(0),
    day(0),
    hour(0),
    minute(0),
    second(0)
  {
  }

  FileTime(struct tm const *);     ///<  FileTime()  construct from a struct tm*

  FileTime(std::string s)          ///<  ls -l time format (mon day year/hour)
  : year(0),
    month(0),
    day(0),
    hour(0),
    minute(0),
    second(0)
  {
    construct_from_string(s);
  }

  bool construct_from_string(std::string s);
    ///<
    /// Construct the time from a unix, ls -l, style string.  Note that there 
    /// is a more flexibile  string to date conversion function, 
    /// @sa parse_date.  That function is meant for user interfaces.
    /// This function is meant for parsing ls -l style output only.
    ///
    /// The construct_from_string method recognizes input of the form:
    ///
    /// @code
    ///   Mon day year_or_hour
    /// @endcode
    ///
    /// For example:
    ///
    /// @code
    ///   Dec 12 2003
    ///   Feb 9 10:45
    /// @endcode
    ///
    /// The difference between the two examples is that instead of supplying a
    /// year as part of the date, the second example supplies the hour of the
    /// day.  This implies that the year is the _current_ year.  When the hour is
    /// missing, then the hour is assumed to be 12:00 am.
    ///
    /// The return value of true means that the string is of an invalid format

  std::string short_string() const;
    ///<
    /// Convert time to a string in the form expected by
    /// construct_from_string() above.
};

//========================================================================================
//========================================================================================
//========================================================================================

struct FileMode
  ///
  /// A representation of a file's mode (ie permissions and type).  If the file is a symbolic
  /// link, the link bit will be 1, but the other bits will refer to file to which the link
  /// points.  Note that this is the design of the FileMode class -- but also note that the
  /// FileMode class does not have way of populating itself.  To obtain a FileMode object's
  /// value, call FileName::file_stat(&fileMode).
  ///
  /// To find the mode of the symbolic link file itself (ie without any reference to its target)
  /// use the FileName::readlink(&mode) function call.  Here's an example use:
  ///
  ///   FileStatus fs;
  ///
  ///   FileName  someFile("someFile");
  ///
  ///   if(someFile.file_stat(&fs))
  ///   {
  ///      // an error has occurred
  ///   }
  ///
  ///   if(fs.mode & FileMode::user_readable)
  ///   {
  ///       // the file's owner (ie user) can read it 
  ///   }
  ///
  /// Note that there is a long list of permission bits to test against.
{
  int mode;  ///< mode The integer representation of the mode, not necessarily the os
             ///  representation!


  int os_mode() const;  ///< return an integer that _is_ the os specific representation
                        ///  of this generic FileMode object.  In the case of symbolic link
                        ///  files, the os_mode() returns the link target's stat::st_mode
                        ///  value.

  operator std::string() const;   ///< convert the permissions to a string

  friend std::ostream& operator<< (std::ostream& o, FileMode const &r)
    ///
    /// operator<<(stream, FileMode const&)
    /// prints the permissions to an output stream
  {
    return o << std::string(r);
  }

  FileMode(): mode(0) {}   ///  FileMode()  default constructor

  FileMode(int m);
    ///
    /// Construct a FileMode from the os' permissions object from an
    /// an integer.
    ///
    /// The permissions integer is assumed to be struct stat's member st_mode

  FileMode(char const*s)
    ///
    /// Construct the mode from a 10 character string formatted like this:
    ///
    ///   drwxrwxrwx
    ///   -rw-rw-rw-
    ///
    /// etc.
  {
    construct_from_string(s);
  }

  FileMode(std::string s)
    ///
    /// Construct the mode from a 10 character std::string formatted
    /// as in the constructor, FileMode(char const*).
  {
    construct_from_string(s.c_str());
  }

  void construct_from_string(char const *string);
    ///<
    ///  Set the mode value given a null terminated character string
    ///  as defined in the constructor, FileMode(char const*).
    ///

  void construct_from_string(std::string const& s)
    ///
    ///  Set the mode value given a null terminated character string
    ///  as defined in the constructor, FileMode(char const*).
    ///
  {
    construct_from_string(s.c_str());
  }


  enum bit_name
    /// names for the various permissions.  Not guaranteed to be portable
  {
       link     = 010000,  ///  symbolic link file
       directory= 004000,  ///  directory  a ubiquitous permission
       set_uid  = 002000,
       set_gid  = 001000,

       user_readable   = 000400,  ///  user_readable  a ubiquitous permission
       user_writeable  = 000200,  ///  user_writeable a ubiquitous permission
       user_executable = 000100,

       group_readable  = 000040,
       group_writeable = 000020,
       group_executable= 000010,

       world_readable   = 000004,
       world_writeable  = 000002,
       world_executable = 0000001
  };

  bool is_dir() const { return mode & directory; }
    ///<
    /// is_dir()
    ///
    /// An inline method indicating that the file is a directory.
    /// 'is_dir' is provided only as an example to see show how the bits
    /// in the permissions work.  For the other bits, you should see is_dir's
    /// implementation and manually access the individual bits.

  bool is_link() const { return mode & link; }
    ///<
    /// An inline method indicating that the file is really a symbolic link that
    /// points to some file somewhere that has the OTHER attributes defined by
    /// the mode member variable.
    ///
    /// The function, FileName::file_stat(), reads through symbolc links to get
    /// the real attributes of the file being pointed to by the link and it is 
    /// these attributes that are described by FileMode::mode.  However, the 
    /// link bit is also set. 
    ///
    /// You can use FileName::readlink() to get the file name referred to by the
    /// link.  It is that file whose attributes are represented by FileMode::mode.
    ///


  void unpackPermissions(std::string &userMode, std::string &groupMode, std::string &worldMode);
      ///<  Split this FileMode object into std::strings representing the permissions.  Each
      ///   permission std::string will look like this:  rwx

  void packPermissions(std::string const &userMode, std::string const &groupMode, std::string const &worldMode);
      ///<  Modify the permissions part of this file mode with data supplied in std::strings.  Each
      ///   mode sub-set (group, user, world) will look like this: rwx
      ///
      ///   Bits in the mode that are not part of the file permissions are not modified by this method

   bool permStringValid(std::string const &string);
       ///<  Return true if the string contains only valid permissions -- permissions must look like this:
       ///     rwx
       ///
       ///   Letters can be left out, upper case is ok, but letters may not be duplicated and no other
       ///   letters, spaces, or -'s can appear
};

//========================================================================================
//========================================================================================
//========================================================================================

struct FileStatus
  ///
  /// The complete description of a file's attributes
  ///
{
   typedef off_t file_size_t;   ///< file_size_t  data type of the 'size' of the file

   FileMode     mode;           ///< the FileMode of the file
   file_size_t  size;           ///< the size of the file
   FileTime     time;           ///< the FileTime of the file
   int          owner_id;       ///< the user id of the file's owner
   int          group_id;       ///< the group id of the file

   FileStatus()
   : mode(),
     size(0),
     time(),
     owner_id(0),
     group_id(0)
   {
       // this constructor is needed to prevent a gcc warning that some members of the
       // class are uninitialized
   }

   std::string convert_to_string(bool resolve_id_names=false) const;
     ///<
     /// std::string convert_to_string(bool resolve_id_names=false) const;
     ///
     /// convert the file's status (ie attributes) into a string --
     /// optionally resolving the user and group id numbers into string
     /// names.  This can be time consuming, so it is optional.

   operator std::string() const
     /// Automatic conversion to a string.  This convert the FileStatus
     /// to a string without mapping id's to names
   {
     return convert_to_string();
   }

   friend std::ostream& operator<< (std::ostream& o, FileStatus const &r)
     /// operator<<(stream, FileStatus const&)
     ///
     /// Output the FileStatus to a stream as a string.
   {
     return o << std::string(r);
   }

   static std::string id2string(int id);
     ///<
     /// id2string(int) use this to convert owner id's to strings
     ///
     /// id2string(int) is a static method that converts user id numbers to
     /// strings.

   static std::string group2string(int id);  // use this to convert group_id to strings
     ///<
     /// group2string(int) use this to convert group id's to strings
     ///
     /// group2string(int) is a static method that converts group id numbers to
     /// strings.

};

//========================================================================================
//========================================================================================
//========================================================================================

class FileContents
  ///
  /// An in-memory representation of a file's contents.
  /// An in memory copy of the file as returned by FileName::slurp().  If the ok() method
  /// returns true then no error occurred during the reading.  However if it returns
  /// false, an explanation of the error can be obtained using 'error()'.
{
private:

   char *begin_;         ///<  start address of the buffer holding the contents
   char *end_;           ///<  end thereof
   char const *error_;   ///<  should be NULL, char* holding the file open error

public:

   FileContents()        /// FileContents()  default constructor
   : begin_(0),
     end_(0),
     error_("not read yet")
   {
   }

   ~FileContents()       ///  destructor
   {
     delete [] begin_;
   }

   typedef char *iterator;               ///<  iterator file data iterator
   typedef char const *const_iterator;   ///<  const_iterator const file data iterator

   const_iterator begin() const { return begin_; }
     ///<
     ///  get an const_iterator to the beginning of the file
     ///

   const_iterator end()   const { return end_;   }
     ///<
     ///  get an const_iterator to the end of the file
     ///

   iterator begin() { return begin_; } ///<  iterator to the beginning of the file
   iterator end()   { return end_;   } ///<  ditto for the end

   void reserve(long bytes) { delete [] begin_; begin_ = new char[bytes]; end_ = begin_ + bytes; }
     ///<
     ///  empty the file and make room for a given size in bytes
     ///

   long size() const { return end_ - begin_; }  ///<  size() how big is the file

   char const *error() const { return error_; }
     ///<
     /// A return of 0 means no error
     ///

   void set_error(char const *msg) { error_ = msg ; }
     ///<
     ///
     /// Force the FileContents to return an error

   void clear_error() { error_ = 0; }
     ///<
     /// Get rid of any previous error

   bool ok() const { return error_ == 0; }
     ///<
     /// See if an error occurred.


};

//========================================================================================
//========================================================================================
//========================================================================================



class FileName
  ///
  /// A file's name and handle for invoking many file system functions.
  ///
  /// FileName is a generalized utility class for performing file operations.
  /// A FileName has much in common with std::string but adds many additional member 
  /// functions that  are associated with operating systems calls involving files:
  ///
  ///   <ul>
  ///          <li> is_dir()
  ///          <li> is_link()
  ///          <li> exists()
  ///          <li> basename()
  ///          <li> dirname()
  ///          <li> convert_to_absolute_path()   -- converts the filename to its full, root relative form
  ///          <li> mkdir()
  ///          <li> and many more.
  ///   </ul>
  ///
  /// Basically this class has everything in it but the kitchen sink.
  ///
  /// A FileName is not a string, but it can be converted to a string &, or string
  /// const & with high speed.
  ///
  ///
{
public:

  struct FileNameSorter
  {
     bool operator()(FileName const &a, FileName const &b) const 
     { 
            int cmp =  StrTool:: compareInsensitive(a,b); 

            if(cmp == 0)
            {
               if(a < b )
                  cmp = -1;
            }


            return cmp < 0;
     }
  };

  typedef std::set<FileName, FileNameSorter> sorted_names_t; ///< Alphabetically sorted set of FileNames.

  typedef std::string              rep_t;          ///<  actual representation of the file's name
  typedef FileStatus::file_size_t  off_t;          ///< The data type of the file's size
  typedef rep_t::const_iterator    const_iterator; ///< const Iterator to characters in the filename
  typedef rep_t::iterator          iterator;       ///< mutable iterator to characters in the file name
  typedef rep_t::value_type        value_type;     ///< needed for CXXTLS_FOREACH

private:
  rep_t             impl_;   ///<  The string that is the file name.

public:

  operator std::string       &()            { return impl_; }         ///<  Convert to string &
  operator std::string const &() const      { return impl_; }         ///<  Convert to const string &
  const_iterator              begin() const { return impl_.begin(); } ///<  Get a pointer to the first char in the name
  const_iterator              end()   const { return impl_.end();   } ///<  Get a ptr to first char after name
  size_t                      size()  const { return impl_.size(); }  ///<  Get length of string
  iterator                    begin()       { return impl_.begin(); } ///<  Get a pointer to the first char in the name
  iterator                    end()         { return impl_.end();   } ///<  Get a ptr to first char after name
  char const *                c_str() const { return impl_.c_str(); } ///<  Get a old fashion C string from the filename
  char const *                data()  const { return impl_.data();  } ///<  Get a pointer to the non-nul terminated filename

  char operator[](size_t i) const { return impl_[i]; }   ///< Get the character at a specified offset from the
                                                         ///  start of the string.

  char &operator[](size_t i) { return impl_[i]; }   ///< Get a ref to the character at a specified offset from the
                                                    ///  start of the string.

  void erase(iterator a)                { impl_.erase(a);    } ///< Erase the character at location
  void erase(iterator a, iterator b)    { impl_.erase(a,b);  } ///< Erase a range of characters
  void erase(size_t o, size_t l)        { impl_.erase(o,l);  } ///< Erase l characters at offset o
  void erase(size_t o)                  { impl_.erase(o);    } ///< Erase the character at offset o
  void insert(iterator a, char c)       { impl_.insert(a,c); } ///< Insert a character at a location
  void insert(size_t o, char const *s)  { impl_.insert(o,s); } ///< Insert a string at an offset
  void insert(size_t o, rep_t const &s) { impl_.insert(o,s); } ///< Insert a string at an offset
  bool empty() const                    { return impl_.empty(); } ///< return true if the filename is an emtpy string

  bool operator<( std::string const &rhs ) const
     ///  Compare this filename's string component against 
  {
    return impl_ < rhs; 
  }

  rep_t substr(size_t s, size_t l) const 
      /// Get a substring given a start ofset a number of characters.
  { 
      return impl_.substr(s,l); 
  } 

  FileName &operator+= (char const *p)  /// append a string to this file name
  {
     std::string *me = &impl_;

     *me += p;

     return *this;
 
  }

  FileName &operator+= (char c) /// append a character to this file name
  {
     std::string *me = &impl_;

     *me += c;

     return *this;

  }


  FileName &operator+= (std::string const &r) /// append a string to this file name
  {
     std::string *me = &impl_;

     *me += r;

     return *this;

  }


  FileName()  /// FileName() default constructor
  {
  }

  FileName(char const *r)  /// construct form a char const*
  : impl_(r)
  {
  }

  FileName(std::string const &r) /// construct from a string
  : impl_(r)
  {
  }

  FileName(FileName const &r) /// copy construct
  : impl_(r)
  {
  }


  FileName expand_tildes() const;
    ///<
    /// @return a copy of self with ~ expansion handled.  For example:
    ///
    ///   ~/.bashrc becomes /home/user/.bashrc 
    ///
    /// and
    ///
    ///   ~lboggs/.bashrc
    ///
    /// becomes
    ///
    ///   /home/lboggs/.bashrc
    ///
    /// Note that the prefix, /home/, is just an example.  The
    /// real results will depend on the home directory of the specified
    /// user -- as determined by passwd() method.
    ///
    /// On Linux/Unix, $HOME defines the home directory.  On windows, a
    /// search for environment variables is done instead:
    ///
    ///   @li %USERPROFILE%
    ///   @li %HOMEPATH%
    ///   @li %HOMEDRIVE%
    ///
    /// If USERPROFILE is found, we use that.  Otherwise, if HOMEPATH is
    /// found, we use that.  Otherwise if HOMEDRIVE is found, we use
    /// %HOMEDRIVE%\.

  rep_t basename(bool remove_extension=false) const;
    ///< Get the filename part of the pathname.  
    ///  @return the file name part of a pathname (including extension)

  rep_t extension() const;
    ///< Get the file name extension part of the pathname.
    ///  @return the file name extension part of a filename.

  rep_t dirname()  const;
    ///< Get only the directory part of the pathname.
    ///  @return  the directory part of self by removing the filename part and leaving the trailing /.

  void  convert_to_absolute_path();
    ///< Adds the name of the current directory to a file name if it has no
    ///  directory part specified.

  void convert_to_DOS_short_path();
    ///<  Converts names with spaces into the shortened form which does not have 
    ///   them.  This is a wrapper around the call, GetShortPathname

  bool  is_syntactically_valid() const;
    ///< @return true if the file name has no bad characters in it.

  bool  exists() const;
    ///< @return true if the filename refers to an existing file.

  bool  is_dir() const;                     ///<  @return true if the named file exists and is a directory
  bool  is_executable() const;              ///<  @return true if the named file exists and is executable
  bool  is_writeable() const;               ///<  @return true if the named file exists and is writeable
  bool  is_readable() const;                ///<  @return true if the named file exists and is readable
  bool  is_createable() const;              ///<  @return true if writeable or parent dir is writeable
  bool  is_absolute_path() const;           ///<  @return true if the filename begins with root directory
  bool  is_root_dir() const;                ///<  @return true if filename is / on unix or c:\ on windows

  FileMode file_mode() const;               ///<  @return FileMode of the named file
  off_t    file_size() const;               ///<  @return the size of named file (or -1 on error)

  bool  file_time(FileTime *t) const;       
    ///<
    ///   Inquire the file time.
    ///
    ///   @param t      A pointer to a FileTime data structure to hold the time.
    ///   @returns      true if an error occurred.

  bool  file_stat(FileStatus* s, bool l=false) const;

    ///< Get status of named file, return true if error.
    ///  If the optional second parameter is 'true'
    ///  then read the status of the link corresponding to
    ///  the file name.  The default is to read through the
    ///  link to get the status of the file or directory to which
    ///  the link points.
    ///
    ///  @param[in] s   The place to store the status of the file whose name
    ///                 is contained in *this.
    ///  @param[in] l   A default flag indicating that if the filename referes to a link
    ///                 the then data for the link should be read instead of
    ///                 for the link itself.  Normally, this is not set, the default
    ///                 is the most desireable form.  Only directory display programs
    ///                 need read the link itself.
    ///  

  bool  matches(char const *pattern) const; 
    ///< Does the filename match a given naming patttern (shell style)
    ///
    ///  @param pattern is a file name search pattern:  "*.c", etc.
    ///
    ///  @return true if the file name matches the pattern.

  bool  matches(rep_t pattern) const        
    /// @copydoc FileName::matches
  { 
    return matches(pattern.c_str()); 
  }


  FileName readlink(int *returnedMode=0) const; 
     ///<  Assume that this filename refers to a symbolic link file
     ///   and attempt to read through the link to get the name of the
     ///   thing it points to.   If the optional mode pointer is passed in
     ///   also read the symbolic link's file mode and return it in
     ///   *returnMode;  Note that normally the FileMode of a symbolic
     ///   link is a mix of the attributes of the link itself and the thing
     ///   it points to.  The mix works like this:
     ///       
     ///   @li  the link bit in the FileMode::mode gets populated with one
     ///        if the file is actually a symbolic link.  
     ///
     ///   @li  The other bits in the mode get populated with the mode of the
     ///        target of the link.
     ///
     ///   If the does not exist or is not a symbolic link, then the
     ///   returned value is empty string.  If the file is a symbolic
     ///   link then an absolute path will be returned.


  bool remove() const;                      
    ///< Remove the file named by this FileName.  @return true on error.

  bool chmod(FileMode) const;
    ///< Set the mode (ie permissions, or attributes) on the file named by
    /// this to the mode specified.
    ///


  bool copy(FileName const &destination) const;
    ///< Copy the file whose name is specified '*this' into a new file
    /// whose name is specified in destination.
    ///
    /// Directories cannot be copied using this function -- and error
    /// will result.
    ///
    /// @param destination The name of the file where the copy goes.
    ///
    /// @return true if an error occurred.

  bool rename(FileName const &destination) const;
    ///< Rename the file whose name is specified '*this' into a new file
    /// whose name is specified in destination.
    ///
    /// @param destination A FileName containing the new name for the file.
    ///
    /// @return true if an error occurred

  bool mkdir(FileMode m=FileMode(" drwxrwxrwx") ) const;
    ///< Make a directory whose name is given by this filename.
    /// Return true if an error occurred.  The 'directory' bit
    /// of the FileMode is assumed -- and thus ignored.
    ///
    /// @return true if an error occurs.

  bool rmdir() const;
    ///< @b Recursively remove the directory named by this filename.
    /// @return true if an error occurred
    ///
    /// If you don't want a recursive delete, check to see if the file has any
    /// members -- use FileName::find_matching("dirname/*", 0x08) and
    /// check for the number of matching results coming back 0.

  void convertToDirectorySearchPattern(std::string const &pattern)
    ///
    /// Assume that this FileName is a directory name and add specified
    /// pattern to it to make it into a directory search pattern (or a simple
    /// filename if the pattern contains no wildcard characters).
    ///
    /// If this is a directory, or it ends in / or \\, add * to the end as
    /// needed to make this a directory search pattern.
    ///
    /// @param pattern is the "*.c", or other pattern you are trying to
    /// construct from this directory.
  {
      convertToDirectorySearchPattern(pattern.c_str());
  }
  
  void convertToDirectorySearchPattern(char const *pattern="*");
    ///< @copydoc FileName::convertToDirectorySearchPattern(std::string const &)


private:

  //
  //  this function does all the work for find_matching().  it exists so that find_matching
  //  can throw an exeption that this guy catches.
  //
  static int find_matching_helper(char const     *pattern, 
                           sorted_names_t *matches, 
                           int             options=0
                          );
public:



  static int find_matching(char const     *pattern, 
                           sorted_names_t *matches, 
                           int             options=0
                          );
    ///<
    /// A static method that finds a list of FileName's matching a specified 
    /// pattern by reading the file system directory and comparing the file
    /// names to the requested pattern.
    ///
    ///  @param pattern is the search pattern:  "*.exe", "/home/dir/files*.*", etc.
    ///  @param matches is a set of strings into which to place the matching filenames.
    ///  @param options is selector for the kinds of files you want to include -- see
    ///         below.
    ///
    ///  @return a count of the matching files.
    ///
    /// @note  The pattern expansion is limited to the final node of any directory
    ///        named in the pattern.  See method, glob(), below if you need full
    ///        bourne shell style patterns.
    ///
    /// The 'options' parameter specifies flags to be interpreted
    /// with respect to the returned list of matching filenames.  The value
    /// passed to options are a bit mask with the following intepretation:
    /// 
    /// @code
    ///   0x00 -- all files are included in the output data but the results
    ///           do not contain the pathnames specified in the pattern, 
    ///           only the basename's of the matches.
    ///
    ///   0x01 -- all files are included in the output data and the results
    ///           _do_ contain the pathname specified in the pattern.
    ///
    ///   0x02 -- only directories are included but no paths are included.
    ///
    ///   0x03 -- only directories, and the paths are included.
    ///
    ///   0x04 -- NO directories are included and no paths are included
    ///   0x05 -- NO directories are included but the paths are.
    ///   0x08 -- Same as 0 but '.' and '..' are not included
    ///   0x10 -- adds case insensitive flag.  Does not apply to character ranges.
    ///   0x20 -- recursive.  If a match is a subdirectory, include it, recursively
    /// @endcode


  static int find_matching(rep_t           pattern, 
                           sorted_names_t* l, 
                           int             options=0
                          )
    /// @copydoc FileName::find_matching
  { 
    return find_matching(pattern.c_str(), l, options); 
  }

  static void glob(std::string const &pattern, std::list<std::string> &matches);
      ///<  A super-duper version of find-matching which lets you use
      ///   patterns like this:
      ///
      ///   @li fred/*/bill/sue/tom/hank/*.c




  void slurp(FileContents *f) const ;
    ///<
    /// Read whole file (binary read) into a FileContents object
    /// and set its error method correctly.

  bool write(std::string const &file_contents, bool binary_mode=true);
    ///<
    /// Write a string to a newly create file whose name is the contained in
    /// this filename.  If the file previously existed, it will be deleted
    /// then recreated.
    ///
    /// @param file_contents is the content of the newly created file 
    /// @param binary_mode is a flag indicating that binary mode will be used
    ///  to write the file (Windows only).
    ///
    /// @return true if an error occurred -- see errno.
    ///
    /// The file will be created using default umask protections and will be written
    /// be written as either ascii or binary base on the binary_mode flag.  On unix
    /// this flag is ignored.  
    /// 

  bool write(char const *begin, char const *end, bool binary_mode=true);
    ///<
    /// Open the file whose name is contained in this FileName object and write
    /// the data found between begin and end to the file in either ascii or binary
    /// mode depending on the binary_mode flag.  On unix, this flag is ignored. The
    /// file will be created with default umask permission.
    /// 
    ///  @param begin is the pointer to the beginning of the data to be written.
    ///  @param end is the end of the buffer to be written
    ///  @param binary_mode (windows only) controls the io mode.
    ///
    //   @return true if a failure occurs.  See errno.
    ///

  size_t slurp(std::list<std::string> *lines) const;
    ///<
    /// Read the whole file (binary read) into a list of strings
    /// return the number of lines read.  This is the fastest way to
    /// read a file into a bunch of strings.
    ///
    /// Ok, so the name is stupid -- what would you have called it?

  struct accumulate_lines
    ///  A class of functor objects
    ///  whose purpose is to provide a generic interface between the
    ///  read_* functions, below, and some container used to hold
    ///  lines read.  
    ///
    ///  An accumulate_lines object has
    ///  two virtual functions, push_back(), and back().  These let
    ///  the read_lines function append new lines to the container
    ///  and access the last line thereof when it needs to update it.
    ///
    ///  See the read() template below for an example use.
  {
    virtual ~accumulate_lines(){}
      ///<
      ///  Destructor must be virtual because other methods are
      ///  virtual.

    virtual void push_back(std::string const &new_line) = 0;
      ///<
      ///  push_back is called to append a new std::string to the
      ///  end of the container

    virtual std::string &back() = 0;
      ///<
      ///  back() is called to get a reference to the last string
      ///  in the container -- presumbaly one created using the above
      ///  push_back method.

    virtual void setBackHasCR(bool flag){ }
      ///<
      /// read_lines() calls this function to indicate to the
      /// line accumulator that the most recently read line, if any
      /// had a CR character removed from it.  This function will only be
      /// called if there is a trailing CR and if the removeCR
      /// flag is passed to read_lines().
           
  };

  std::string read_lines(accumulate_lines &accum, size_t *line_count_ptr=0, bool removeCR=false);
    ///<
    ///  This function lets you read lines from the file whose name
    ///  contained in this filename very quickly.  You probably want to
    ///  use the FileName::read_file function below instead of calling
    ///  this function.  It simplifies the interface significantly.
    ///
    ///  @param accum is an container-wrapper that will hold the input data.
    ///  @param line_count_ptr is an optional pointer to a size_t holding the count of
    ///         read lines.
    ///  @param removeCR is an optional boolean that tells the function to
    ///         remove a trailing \\r character, i
    ///
    ///  @return a string containing the strerror() contents or "" if no errors.
    ///
    ///  You must create a
    ///  FilenName::accumulate_lines object to serve as an interface
    ///  to the container into which you are storing the read lines.
    ///  The lines will be read in binary mode.  The line separator, \\n,
    ///  will not be included in the strings returned -- nor will any
    ///  single trailing \\r (On unix you must set the
    ///  removeCR flag to get the trailing \\r removed.  On windows, this
    ///  is neither necessary nor helpful.)
    ///
    ///  This function returns a std::string contain any error message
    ///  that might have occurred.  If the return value is empty, then
    ///  no error occurred.
    ///
    ///  Do not attempt to use the push_back method to do anything but
    ///  accumulate lines.  The read_lines algorithm handles very long
    ///  lines, but it has a fixed size internal buffer.  As the data is
    ///  read from the file it is stored in this internal buffer.  When a
    ///  line exceeds the length of this buffer, partial lines will be
    ///  pushed into the  the container.  The back() function will
    ///  subsequently be called to get access to the most recently pushed
    ///  partial line so that the partial line can be completed using the
    ///  std::string::append method.
    ///
    ///  See the body of FileName::read() below for an example of
    ///  using read_lines().

  template<class Container>
  struct Accum
  : public accumulate_lines
    ///
    ///  An Accum object is a template class object whose role
    ///  is to provide a generic interface for any standard sequence
    ///  container to the accumulate_lines virtual interface.
    ///
    ///  By rights, this should be a nested class found within the
    ///  read_file template function below, but you can't instantiate
    ///  a template given a function local class.  Sooner or later these
    ///  silly restrictions will go away, but now they haven't.
    ///
    {
      Container& container_;

      Accum(Container& container)
      : container_(container)
      {
      }

      void push_back(std::string const &s) { container_.push_back(s); }
      std::string & back()                 { return container_.back();}

    };

  template<class Container>
  std::string read_file(Container &container, size_t* line_count_ptr=0)
    ///
    ///  Read a whole file into a container of std::string's.  Return any
    ///  error text as a std::string as well.  If the return string is
    ///  empty then no errors have occurred.
    ///
    ///  @param container can be any sequence container that should holds lines
    ///         of text from the input file (as std::string)
    ///
    ///  @param line_count_ptr is a pointer to the location where the count of
    ///         read lines will be stored.
    ///
    ///  @return errorString is a string describing any problems you might have
    ///          encountered.
    ///
    ///
    ///  This function differs from the FileName::slurp() method in
    ///  that slurp reads the entire file into a single giant array of
    ///  characters.  This function reads it into a any container of std::strings
    ///  which you define.  Note that on large files, vector<std::string>
    ///  gives you O(N) inserts whereast list<std::string> gives you O(1)
    ///  inserts.  A skiprope<std::string> would give you O(ln(N)) inserts
    ///  as well as O(ln(N)) array indexing -- unlike the list which has none
    ///  what so ever.
    ///
  {

    Accum<Container> accum(container);

    return read_lines( accum, line_count_ptr );

  }

  static FileName findTempDir();  
     ///<  Return the name of a temporary directory for use in
     ///   in writing temporary files.  On unix/linux, you get
     ///   "/tmp/" returned.  On Windows, you get one of several directories
     ///   include some that might be created by this function. 
     ///
     ///   Note that in all cases, the directory name ends in /.
     ///
     ///   Note also, that the current directory will never be returned.


  static FileName mktemp(FileName dir="./");
    ///<
    /// static function that will return a FileName with the name of a temporary file in the
    /// specified directory.
    ///
    ///  @param dir is the name of a directory where the temporary file will go.
    ///
    ///  @return a FileName object with the name of a file that does not yet exist
    ///   in the specified directory.  The name will be some numeric pattern.
    /// 
    /// That is, this function produces the name of a file in the specified directory
    /// that does not yet exist -- for use in creating temporary files.
    /// The directory parameter is optional -- the default being the
    /// current directory.
    ///
    /// When specifying a directory name, you \b must supply the trailing
    /// slash and you may also include a prefix for the file.  Here are
    /// some valid examples:
    ///
    /// @code
    ///   FileName::mktemp();
    ///   FileName::mktemp("/tmp/");
    ///   FileName::mktemp("../../junk_");
    /// @endcode


  static char PATH_separator;  ///< Either : or ; depending on the os
                               ///  This variable defines what character
                               ///  separates components of the PATH environment
                               ///  variable.
                        
  struct is_separator
    ///
    ///  A functor that tells you if a specifed character
    ///  is a path sepator.  It is meant to be used as a parameter to
    ///  StrTool::parse_words -- or any function needing a predicate indicating
    ///  that a character is the path separator.
    ///
  {
    bool operator() (char c) const
    {
      return c == FileName::PATH_separator;
    }
  };

  static FileName find_executable(std::string progname, std::string path="");
    ///<
    /// Search the specified path, or the current PATH environment variable
    /// for the named program.  
    ///
    ///   @param progname is the name of an executable to search for
    ///   @param path is a platform specific directory path (unix:  colon
    ///               separated, windows: semicolon separated)
    ///
    ///   @return the name of the found file -- or empty string.
    ///
    /// How this works varies a bit from os to os.
    /// If no path is specified, or the path is an empty string, the current
    /// PATH environment variable will be used for the search.  On Windows,
    /// the PATH variable is separated by spaces and semicolons but on unix
    /// it is separated by only colons.  Further, on Windows, the
    /// program name may not include the file name extension which is one
    /// of .bat, .com, .exe, .cmd.  On these systems, this will be automatically
    /// appended during the search.
    ///

  static FileName find_file_in_path(std::string filename, std::string path="");
    ///<
    ///  Search a path for a given file -- executable or not.  See find_executable
    ///  for the general idea.  This function, finds any file, not just executables.
    ///
    ///   @param filename is the name of the file to search for
    ///   @param path is a directory search path dependend in form on the os you are on.
    ///
    ///   @return the name of the found file -- or empty string.
    ///  

  std::string shorten(int max_chars) const;
    ///<
    ///  Produce a string which is a shortened version of the filename by
    ///  deleting excess characters and inserting an elipses in the middle to
    ///  signify missing characters.  The name will not refer to a real file (unless
    ///  by accident).

  FileName operator+(FileName const &r ) const /// append a file name to this file name
  {
    std::string const &me = *this;
    std::string const &other = r;


    return FileName(me + other);
  }

  FileName operator+(char const* r) const
    ///  concatenate the filename with a c style string to make a new filename
  {
    std::string const &me = *this;


    return FileName(me + r);
  }

  FileName operator+(std::string const &r) const
    ///  concatenate the filename with a std::string to make a new filename
  {
    std::string const &me = *this;


    return FileName(me + r);
  }

  template<class T>
  FileName &operator=(T const &rhs)
    /// Assign this FileName from some other string source.
  {
     impl_ = rhs;

     return *this;
  }



  friend std::string operator+(std::string const &lhs, FileName const&rhs)
    ///  global free function to concatenate strings and filenames, producing
    ///  std::strings
  {
    return lhs + (std::string const &)(rhs);
  }

  friend std::string operator+(char const *lhs, FileName const&rhs)
    ///  global free function to concatenate c style strings and filenames
    ///  producing std::strings'
  {
    return lhs + (std::string const &)(rhs);
  }

  static int compareIgnoringPathSeparatorDifferences(char const *l,
                                                     char const *r,
                                                     size_t      lsize,
                                                     size_t      rsize
                                                    )
      ///
      ///  @return a signed integer indicating the comparison of l and r
      ///          while ignoring / vs \ differences.  negative means
      ///          that l is less than r, positive means that it is greater
  {


     size_t smaller = lsize < rsize ? lsize : rsize;

     for(size_t i = 0; i < smaller; ++i)
     {
         char lc = l[i];
         char rc = r[i];

         if(lc == '\\')
            lc = '/';

         if(rc == '\\')
           rc = '/';

         if(lc < rc)
            return -1;

         if(lc > rc)
            return 1;

         // the characters are equal, so keep comparing.


     }

     // if we get here, the two strings are euqal up in the range [0..smaller-1].
     // which represents the common sub-range of the the two strings.

     if(lsize < rsize)
        return -1;
     else
     if(lsize > rsize)
       return 1;

     return 0; // strings are equal

      
  }

  static int compareIgnoringPathSeparatorDifferences(std::string const &l,
                                                     std::string const &r
                                                    )
      ///
      ///  @return a signed integer indicating the comparison of l and r
      ///          while ignoring / vs \ differences.  negative means
      ///          that l is less than r, positive means that it is greater
  {
      return compareIgnoringPathSeparatorDifferences(l.data(), r.data(), l.size(), r.size());
      
  }



  bool operator==(FileName const rhs) const
    ///  Compare two filenames
  {
    int cmp = compareIgnoringPathSeparatorDifferences(*this, rhs);

    if(cmp != 0)
       return false;

    return true;


    // return ((std::string const &)(*this)) == (std::string const&)(rhs);
  }

  bool operator==(std::string const rhs) const
    /// compare a filename to a string
  {

    int cmp = compareIgnoringPathSeparatorDifferences(this->data(), rhs.data(), this->size(), rhs.size());

    if(cmp != 0)
       return false;

    return true;

    // return ((std::string const &)(*this)) == rhs;
  }

  bool operator==(char const *rhs) const
    /// compare a filename to a c style string
  {
    int cmp = compareIgnoringPathSeparatorDifferences(this->data(), rhs, this->size(), strlen(rhs));

    if(cmp != 0)
       return false;

    return true;

    
    return ((std::string const &)(*this)) == rhs;
  }


  friend bool operator==(std::string const &lhs, FileName const& rhs)
    ///  global free function to compare a std::string to a filename
  {

    int cmp = compareIgnoringPathSeparatorDifferences(lhs, rhs);

    if(cmp != 0)
       return false;

    return true;

    // return lhs == (std::string const &)(rhs);
  }

  friend bool operator==(char const*lhs, FileName const& rhs)
    ///  global free function to compare a c style string to a filename
  {

    int cmp = compareIgnoringPathSeparatorDifferences(lhs, rhs.data(), strlen(lhs), rhs.size());

    if(cmp != 0)
       return false;

    return true;
    
    // return lhs == (std::string const &)(rhs);
  }

  static FileName getcwd(); ///<  static function to get the current directory as a filename

  class Tree
    ///
    ///  A tree of file names, representing a directory hierarcy.
    ///
    ///  The FileName::Tree class is meant for implementing simplistic
    ///  directory hierarchies.  It is not a general purpose tool.  See
    ///  FileName::tree();
  {
  public:
     typedef std::list<Tree*> children_t; ///< data storage type for the tree

     friend class FileName;  

  private:

     FileStatus      status_;   ///<  This node's status
     std::string     name_;     ///<  This node's name
     children_t      children_; ///<  This node's children (if this is a dir)

     Tree*           parent_;   ///<  This node's parent directory

     int             depth_;    ///< Depth within the hierarchy, top = 0

  public:

      Tree(Tree* parent=0) 
      : parent_(parent),
        depth_(0)
      {
        ///  Construct an empty tree
      } 

     ~Tree();               ///<   destructor
      Tree(Tree const &r);  ///<   copy constructor

      Tree(std::string const &name, FileStatus const &status, Tree* parent=0, int depth=0)
      : status_(status),
        name_(name),
        parent_(parent),
        depth_(0)
      {
        ///   Construct a tree node with no children but specified
        ///   name and file info
      }

      Tree &operator=(Tree const &r)
      {
        this->~Tree();       // delete self
        new (this) Tree(r);  // copy construct from r
        return *this;
      }


      FileStatus const&  status() const   { return status_; }
        ///<  get info for this node

      FileName           name() const     { return name_;   }
        ///<  get name of this node

      FileName           path() const;
        ///<
        ///  get pathname of this node -- as specified by parent hierarchy
        ///  not by inquiring the current directory -- use convert_to_absolute_path
        ///  if you need the complete pathname.

      int                depth() const  { return depth_; }
        ///<
        ///   Return the relative depth in of this node within the tree.  The top
        ///   item in the tree has depth of 0, its children have depth of 1, etc

      children_t  const&  children() const { return children_; }
        ///<  get children of this node

      void add_child(FileName const &name, FileStatus const &status)
        ///  add child to this node (no error detection is done)
      {
          children_.push_back(new Tree(name,status,this,this->depth_+1));
      }

      void add_child(Tree *tree)
        ///  add a child to the tree and give ownership thereof to the tree 
      {
        tree->parent_ = this;
        tree->depth_  = depth_+1;
        children_.push_back(tree);
      }

      void remove_child(FileName const &name);
        ///<  remove a specified child of this node ( warning:  O(n)!)

  };

  std::auto_ptr<Tree> tree(bool only_dirs=false, bool allFiles=false, bool debug=false) const;
                           ///<  Get the subdirectory tree starting with
                           ///   this filename.  If this filename refers to
                           ///   a normal file, then tree will have only 
                           ///   one node. If this does not refer to a valid
                           ///   file, one node will be returned but the
                           ///   'status' will indicate that it is not a good
                           ///   file.  If this FileName is a directory, then
                           ///   the sub-directory tree will be returned.  If
                           ///   any node's status is not avaialable, its status
                           ///   will indicate that. 
                           ///
                           ///   @param only_dirs True implies that only directories
                           ///                    should be returned.  False by default.
                           ///
                           ///   @param allFiles  True implies that files beginning with  .
                           ///                    will be included.  False by default.
                           ///
                           ///   @param debug     True implies that debugging information
                           ///                    should be written to stdout.

private:
    std::auto_ptr<Tree> 
    tree_helper(std::string const &path, bool only_dirs=false, int depth=0, bool allFiles=false, bool debug=false) const;
        ///<   Used by tree() to recurse through directory hierarchies.
        ///    
        ///    @param path The name of the directory to recurse through.
        ///    @param only_dirs  True if only directories should be included.
        ///    @param depth      The maximum depth of the recursive directory search.
        ///    @param allFiles   True if all files should be included in the results (.* files)
        ///    @param debug      True if debug information should be printed to cout.
};

//========================================================================================
//========================================================================================
//========================================================================================

inline std::ostream &operator<< (std::ostream &s, FileName const &rhs)
{
    std::string const &data = rhs;

    return s << data;
}

//========================================================================================
//========================================================================================
//========================================================================================
inline bool operator!= (FileName const &lhs, std::string const &rhs)
{
    std::string const &data = lhs;
    std::string const &rdata= rhs;

    return data != rdata;

}

//========================================================================================
//========================================================================================
//========================================================================================
inline bool operator!= (std::string const &lhs, FileName const &rhs)
{
    std::string const &ldata = lhs;
    std::string const &rdata = rhs;

    return ldata != rdata;

}



} // namespace cxxtls

#endif
