//
// Copyright 2002-2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

 #include <cxxtls/thread.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/foreach.h>

#include <map>
#include <utility>
#include <memory>
#include <time.h>

namespace cxxtls
{

///@file threadPoolPage
///
/// This file defines classes and functions that implement command handling thread pools.
/// @see threadPage
///
/// In particular, see cxxtls::ThreadCommandHandlerPool and cxxtls::ThreadCommandHandler


template<class CommandObject>
struct ThreadCommandHandler
    ///   A threaded object that executes commands.   
    ///
    ///   Derive from this class and superseed the handle() method to implement
    ///   your thread pool's behavior.
{

    virtual bool handle(CommandObject &cmd) const = 0;
        ///<  Handle the specified command and return true if the handler
        ///   decides that the calling thread should terminate.  Mostly,
        ///   the return value will be false.
        ///
        ///   @param cmd  The command to execute
        ///   @return true if the calling thread should terminate.
        ///
        ///   This function will be called simultaneously by multiple threads.
        ///
        ///   Thread local storage is not directly available to this handler.
        ///           so you can't expect to keep some sort of state associated with each
        ///           thread in this handler.   This is @b super @b important!  Do not
        ///           make any assumptions about the chronological order of commands being
        ///           processed by this handler.  
        ///
        ///   The commands themselves should contain any state information needed by the handlers.
        ///
        ///   

     virtual ~ThreadCommandHandler()
     {
         CXXTLS_HEAVYDEBUG("~ThreadCommandHandler()");
     }

};  //  ThreadCommandHandler


struct ThreadCommandHandler_base
   /// base class from which the ThreadCommandHandler template derives.  This class
   /// exists so that CommandHandlerThread's can signal to the thread pool that they
   /// are dying.
{
    virtual void deleteWorker(int threadNumber) = 0;  ///< handle the death of a worker.
};


template<class CommandObject>
struct CommandHandlerThread
    ///
    ///   An object whose run() method performs the acto of reading commands from a queue
    ///   and executing them.
    ///
    ///   One object of this type will be created for each thread in a thread pool.  However,
    ///   all threads share the same command handler object.  The @ref run() method invokes
    ///   the @ref handle() method on the ThreadCommandHandler<CommandObject> shared by the
    ///   threads in the pool.
    ///
    ///   Note that CommandHandlerThreads fundamentally cannot have any "state".  The command
    ///   objects themselves must then contain any state information needed.
    ///
{
    ThreadCommandHandler<CommandObject> const  *handler_;
    ThreadQueue<CommandObject>                 *queue_;
    ThreadCommandHandler_base                  *pool_;
    int                                         id_;

    CommandHandlerThread(ThreadCommandHandler<CommandObject>  const *handler,
                         ThreadQueue<CommandObject>                 *que,
                         ThreadCommandHandler_base                  *pool,
                         int                                         id
                        )
    :  handler_(handler)
    ,  queue_(que)
    ,  pool_(pool)
    ,  id_(id)
    {
    }

    void run()
       ///  Read and execute commands until the input queue is terminated or
       ///  until the handler function for a command says to.
       ///
    {
        CXXTLS_HEAVYDEBUG("Running thread " << id_);

        for(;;)
        {
            CommandObject tmp;
            
            if(!queue_->waitPop(tmp))
            {
               CXXTLS_HEAVYDEBUG("Thread " << id_ << " has noticed thread queue termination ");
               break;  // the queue has been terminated
            }

            CXXTLS_HEAVYDEBUG("Thread " << id_ << " has received " << tmp);

            if(handler_->handle(tmp))
               break;  // the handler says that this thread should terminate
        }

        pool_->deleteWorker(id_);  // signal my death
    }

    

}; //  CommandHandlerThread<CommandObject>

template<class CommandObject>
struct ThreadCommandHandlerPool
: public ThreadCommandHandler_base
    ///  A pool of threads designed to handle commands of a particular type which is defined by the template parameter, CommandObject.
    ///  The commands that are handled must be inserted into the pool's input queue using the push() method.  There is @b no
    ///  @b ordering @b guarantee with regards to the commands being handled by this pool.  
    ///
    ///  Note that the destructor of the ThreadCommandHandlerPool kills off all its worker threads, so
    ///  when developing an application with a ThreadCommandHandlerPool, you must make sure to destroy it, but
    ///  not so soon that the handler threads are still working on data.  One way to do that is call the member
    ///  function waitForShutdown().  This function will suspend the current thread until one of the following
    ///  things happens:
    ///
    ///    @li all of the individual worker threads terminate because the handler functions have encountered commands
    ///        that tell them to terminate.
    ///
    ///    @li some other thread has invoked the shutdown() method.
    ///
    ///  When you construct a pool, you must specify a pointer to a command handler object that will exist as
    ///  long as the pool itself exists.  You must also tell it how many threads to use.   The command handler object
    ///  will be used by the worker threads, in a shared fashion, to handle the commands that are sent to this pool.
    ///
    ///  It is tempting to use a lot of threads, but for heavily compute intensive applications, you should 
    ///  not use more than 1 + the number CPUs for the thread count.  Threading will actually make things slower
    ///  for compute intensive applications.  But if your threads are waiting in I/O a  a lot, then you can use 
    ///  more threads.  The I/O impact will swamp the thread based performance degradation.
    ///
    ///  See ThreadCommandHandler for caveats about writing command handlers.
    ///
    ///  Note that the command handlers are run in parallel, so you can't rely on the cronological order of
    ///  insertion of the commands into the queue_ to arrange program order.
    ///
    ///  Be advised.  Be VERY advised.
    ///
    ///  Standard usage pattern.
    ///
    ///  @li Decide on an application design that can benefit from mulithreaded command handling.  Not all can.
    ///  @li Define a class that handles the execution of commands.  Remember to make the handler thread safe.
    ///  @li Declare a command handler variable which will exist for at least as long as the thread pool.
    ///  @li Declare a thread pool in the main thread.  Pass it the address of the command handler and the count of threads to use.
    ///  @li Call waitForShutdown().  Note that many threads can wait on the same thread pool to shut down.
    ///  @li Destruct the thread pool.
    ///  @li destruct the command handler variable.
    ///  
{
    Mutex                                                 guardian_;  ///<  the mutex guarding modifications to the pool itself.

    ThreadCommandHandler<CommandObject>            const *handler_;   ///< calling thread owns the handler object
    ThreadQueue<CommandObject>                            queue_;

    typedef std::map<int, ThreadId>                       ThreadInfo;
    ThreadInfo                                            threads_;   ///< mapping thread number to thread id.  

    ConditionVariable                                     shutdownGate_;  ///< used to orchestrate a wait for all threads to shut down

private:

    struct threadparm_t
       /// The parameters passed to each thread on startup.  Objects of this
       /// type are created using operator new and passed to worker threads.  They must be deleted by 
       /// the thread that that consumes them.
    {
        ThreadCommandHandler<CommandObject> const *handler_;
        ThreadQueue<CommandObject>                *queue_;
        ThreadCommandHandler_base                 *pool_;
        int                                        id_;

        threadparm_t(ThreadCommandHandler<CommandObject> const *handler,
                     ThreadQueue<CommandObject>                *que,
                     ThreadCommandHandler_base                 *pool,
                     int                                        id
                    )
        :  handler_(handler)
        ,  queue_(que)
        ,  pool_(pool)
        ,  id_(id)
        {
        }

    };

public:

    struct InvalidThreadParmsException{};
    struct InvalidWorkerThreadIdException{};

private:

    static CXXTLS_DECLARE_THREAD_FUNCTION(runThread, threadParm)
       /// Run the thread.  This function assumes threadParm was new'd up and that we 
       /// should delete it on exit. 
    {
        threadparm_t *myParms = reinterpret_cast<threadparm_t*>(threadParm);

        if(!myParms)
           throw InvalidThreadParmsException();

        CommandHandlerThread<CommandObject> executor(myParms->handler_, myParms->queue_, myParms->pool_, myParms->id_);

        executor.run();

        delete myParms;

        return false;
    }

    void deleteWorker(int threadNumber)
        ///  A callback that the worker threads invoke to signal that they are going bye bye.
    {
       ScopedMutexLock lock(guardian_);

       CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.deleteWorker(" << threadNumber << ")");

       ThreadInfo::iterator where = threads_.find(threadNumber);

       if(where == threads_.end())
          throw InvalidWorkerThreadIdException();

       threads_.erase(where);
    }

public:

    ThreadCommandHandlerPool( ThreadCommandHandler<CommandObject> const *handler,
                              int                                        threads
                            )
       ///  Construct a thread command handler pool with a specified number of threads
       ///  using a specified command handler object to execute the commands being
       ///  sent to the threads.  The handler is shared among all threads and must continue
       ///  to exist during the lifetime of the pool.  It is not owned by this
       ///  pool, so make sure you create/delete it as you create/delete the pool.

    :  handler_(handler)
    ,  queue_(true) // multiple readers
    {
        ScopedMutexLock lock(guardian_);  /// make sure that deleteWorker() is not called during
                                          /// the construction phase

        while(threads)
        {
           ThreadId id = 
           createThread(&ThreadCommandHandlerPool<CommandObject>::runThread, 
                        new threadparm_t(handler_,
                                         &queue_,
                                         this,
                                         threads
                                        )
                      );
           threads_.insert( std::pair<int,ThreadId>(threads, id) );

           --threads;
        }
    }

    ~ThreadCommandHandlerPool()
    {
        CXXTLS_HEAVYDEBUG("~ThreadCommandHandlerPool()");

        shutdown();
       
    }

    void push(CommandObject const &cmd)
    {
       CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.push(" << cmd << ")");
       queue_.push(cmd);  // queue_ is thread safe by default
    }

    bool empty() const
       ///<  determine if the input queue for all the worker threads is empty.
       ///
       /// @note  This is an instanteous value that can change quickly, so it is
       ///        mainly interesting only for debugging.
    {
       return queue_.empty(); // queue_ is by definition thread safe, so we don't have worry about locking
    }

    bool terminated() const 
       /// return true if the pool is being shutdown.
    {
       return queue_.inputTerminated();
    }

    void shutdown()
        ///  Shutdown of the worker threads.
        ///  
        ///  This is accomplished by setting the terminateInput flag on the queue used by the
        ///  push() method to insert commands for the worker threads to execute.
        ///
        ///  Note that shutdown is not immediate.  Worker thereads will not terminate until
        ///  their current task is complete.
        ///
        ///  Warning:  A worker thread cannot successfully perform a shutdown.  Instead, they should
        ///  communicate to some other thread (such as the main thread) the fact that a shutdown is
        ///  required and that thread should call shutdown.  Here's the easy way:
        ///
        ///  @li declare a commonly accessible variable that is a threadQueue<int>, this is the shutdown queue
        ///  @li the main thread creates a thread pool and pushes a command into the pool's input
        ///      queue
        ///  @li the main thread waits on the shutdown queue.  When the data arives, call shutdown()
        ///  @li if a thread determines that shutdown is required, it should push an int into the
        ///      shutdown queue, then return as normal.
    {

        CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.shutting down");

        queue_.terminateInput();  // this will terminate all command handler threads

        // The above call will cause the worker threads to invoke deleteWorker(), and thus
        // remove themselves from the list of running threads.

        for(;;)
        {
            bool empty = false;

            {
               ScopedMutexLock lock(guardian_);

               empty = threads_.empty();
            }

            if(empty)
               break;
            else
            {
                // sleep 1 ms so as not to burn up all the cpu

#ifndef _MSC_VER
                timespec one_millisecond;

                one_millisecond.tv_sec  = 0;
                one_millisecond.tv_nsec = 1000 * 1000;

                nanosleep(&one_millisecond,0);
#else
                Sleep(1);
#endif
            }
        }


        CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.shutting down finished, signalling waitors");

        ScopedMutexLock lock(guardian_);

        CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.notifying waitors()");

        shutdownGate_.notifyAll();

    }

    void waitForShutdown()
       /// wait until some other thread calls the shutdown() function on the threadpool.
    {
       CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.waiting for shutdown()");

       ScopedMutexLock lock(guardian_);

       shutdownGate_.wait(lock);

       CXXTLS_HEAVYDEBUG("ThreadCommandHandlerPool<>.shutdown detected.  Kansas has gone bye bye");

    }

}; // ThreadCommandHandlerPool


}  // namespace cxxtls
