//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
#include <iostream>
#include <cxxtls/thread.h>
#include <set>
#include <cxxtls/foreach.h>
#include <cxxtls/lastErrno.h>
#include <cxxtls/ScriptLineStream.h>
#include <fcntl.h>

#ifndef _MSC_VER
#include <unistd.h>
#endif

using namespace std;


namespace cxxtls
{
    ScriptLineStream::
   ~ScriptLineStream()
   {   
   
      // never close the string stream, if used, because we do not own it.

       
       if(closeAtExit_)
       {
#ifdef _MSC_VER
          CloseHandle(fh_);
#else
          ::close(fh_);
#endif
       }

       console_.reset(0);  // force the destruction of the command console before
                           // the automatic destructors fire because the console has an internal
                           // reference to the history_ variable and its destructor must run 
                           // before the destruction of the history_ variable.


   }
    
    
    ScriptLineStream::
    ScriptLineStream(string const &name, string const &prompt)
    :  stringFile_(0)
    {
        setDefaults();   

#ifdef _MSC_VER

        handle_t hFile;

        string filteredName;

        CXXTLS_FOREACH(char c, name)
        {
           if(c == '/')
              c = '\\';

           filteredName += c;
        }

        SetLastError(0);


        hFile = CreateFile(    filteredName.c_str(),   // name of file to read
                               GENERIC_READ,           // open for reading
                               FILE_SHARE_READ,        // let others read too
                               NULL,                   // child processes need this handle
                               OPEN_EXISTING,          // 
                               FILE_ATTRIBUTE_NORMAL,  // normal file
                               NULL                    // no attr. template
                          );
           
        if (hFile == INVALID_HANDLE_VALUE) 
        {
            cerr << filteredName << ":1: Error opening file, " << cxxtls::lastErrno() << endl;
            return;
        }

        DWORD consoleMode;


        if(GetConsoleMode(hFile, &consoleMode))
        {
           isConsole_ = true; 
        }



#else
       fh_ = open(name.c_str(), O_RDONLY);

       if(fh_ < 0)
       {
            cerr << "Unable to open file:  " << name << endl;
            return;
       }

       if(isatty(fh_))
           isConsole_ = true;

#endif

        fh_          = hFile;
        good_        = true;
        closeAtExit_ = true;
        filename_    = name;
        error_       = false;

        promptStack_.push_back(prompt);
    }

    ScriptLineStream::
    ScriptLineStream(ScriptLineStream::handle_t  handle,
               bool                              closeAtExit,
               string const &                    prompt
              )
    :  stringFile_(0)
    {
        setDefaults();   
        
        fh_          = handle;
        closeAtExit_ = closeAtExit;
        good_        = true;
        error_       = false;

        promptStack_.push_back(prompt);


#ifdef _MSC_VER
        DWORD consoleMode;


        if(GetConsoleMode(handle, &consoleMode))
        {
           isConsole_ = true; 
        }
#else
        if(isatty(fh_))
           isConsole_ = true;
#endif


        if(isConsole_)
        {
            openCommandConsole();
        }


    }

    void 
    ScriptLineStream::
    openCommandConsole()
    {
            string historyFile;

#ifdef _MSC_VER
            char const *homevar = getenv("USERPROFILE");

            if(homevar == 0)
               historyFile=".commandhistory.txt";
            else
            {
                historyFile = homevar;
                historyFile += "/.commandHistory.txt";
            }
#else
            char const *homevar = getenv("HOME");

            if(homevar == 0)
               historyFile=".history.txt";
            else
            {
                historyFile = homevar;
                historyFile += "/.commandHistory.txt";
            }
#endif
            console_.reset( CommandConsole::getDefaultConsole(historyFile, history_) );
            workingDirectory_ = ".";
    }


    ScriptLineStream::
    ScriptLineStream(string const &        prompt)
    :  stringFile_(0)
    {
        setDefaults();   

#ifdef _MSC_VER
        fh_          = GetStdHandle(STD_INPUT_HANDLE);;
#else
        fh_          = 0;
#endif

        closeAtExit_ = false;
        good_        = true;
        error_       = false;

        promptStack_.push_back(prompt);

#ifdef _MSC_VER
        DWORD consoleMode;


        if(GetConsoleMode(fh_, &consoleMode))
        {
           isConsole_ = true; 
        }
#else
        if(isatty(fh_))
           isConsole_ = true;
#endif

        if(isConsole_)
        {
            openCommandConsole();
        }

    }

    ScriptLineStream::
    ScriptLineStream(std::stringstream &input)
    :  stringFile_(&input)
    {
        setDefaults();   

        fh_ = cxxtls::ScriptLineStream::handle_t(-1);

        closeAtExit_ = false;
        good_        = stringFile_->good();
        error_       = false;

        promptStack_.push_back("");

        isConsole_ = false;
    }


    bool
    ScriptLineStream::
    readFile(ScriptInputLine &text)
    {

      text.text_.resize(0);

       if(isConsole_)
          return false;   // this makes no sense

       char buffer[32768];

       int bytesRead;

       for(;;)
       {
           // read the whole file into the text.text_ member

           if(stringEofEncountered_)
           {
               good_ = false;         // read past eof
               bytesRead = 0;
               return !text.text_.empty();
           }
           
           if(stringFile_)
           {
               
               stringFile_->read(buffer, sizeof(buffer));
               
               bytesRead = stringFile_->gcount();  // count of bytes read during last operation
               
               if(bytesRead == 0)
               {
                  stringEofEncountered_ = true;

                  return !text.text_.empty();
               }
               else
               {
                   text.text_.append(buffer, buffer + bytesRead);
               }
           }
           else
           {
              // normal file read operations

              // read one block from the file

#ifdef _MSC_VER
              DWORD inputBytes=0;
              
              SetLastError(0);
  
              if(!ReadFile(
                            fh_,
                            buffer_,
                            sizeof(buffer_),
                            &inputBytes,
                            0
                           )
              )
              {
                 // io error means return.
  
                  DWORD err = GetLastError();
  
                  CXXTLS_LOG("ReadFile error:  " << cxxtls::lastErrno());
  
                  stringEofEncountered_ = true;


                  good_ = false;
                  error_ = true;
  
                 if(text.text_.empty())
                    return false;
  
                 return true;   // if we get here, the text.text_ buffer has no data in it
              }
  
              if(inputBytes == 0)
              {
                 good_ = false;
  
                 return !text.text_.empty();
              }
  
              bytesRead = inputBytes;

              text.text_.append(buffer, buffer + bytesRead);

              return true;

#else 
              bytesRead = read(fh_, buffer, sizeof(buffer));
              
              if(bytesRead <= 0)
              {
                 good_ = false;
              
                 error_ = int(bytesRead) < 0;
              
                 stringEofEncountered_ = true;

                 return !text.text_.empty();
              }
#endif


           }
       } // read the whole file

    }  // read the whole file



    bool 
    ScriptLineStream::
    readLine(ScriptInputLine &text)
    {
        text.text_.resize(0);  // make sure its empty if eof occurs

        if(isConsole_)
        {
            cout << getPrompt();

            text.text_ = "";

            int lineNumber=0;

            text.filename_ = "console";  


            bool eof = !console_->readline(text.text_, lineNumber, workingDirectory_);


            if(!eof)
            {
               text.lineNumber_ = lineNumber;

               if(recordHistory_)
               {
                   lineGroup_.push_back(text.text_);
               }
               return true;
            }
            else
                return false;
        }


        text.filename_     = filename_;

        if(good_)
        {
            ++lineNumber_;
        }

        text.lineNumber_   = lineNumber_;

         text.text_.resize(0);

        if(bytesInBuffer_)
        {
           // there are remaining bytes in the buffer_, left over from the last read operation.

           char const *where = (char*)memchr(buffer_, '\n', bytesInBuffer_);

           if(where)
           {
               // there is a newline in the data in the buffer, lets return that
               // and remove those bytes from the buffer.

               text.text_.append(buffer_, where+1);   // use the data, include the \n in the data

               memcpy(buffer_, where+1, bytesInBuffer_ - (where-buffer_) - 1);  

               bytesInBuffer_ -= 1 + (where - buffer_);

               if(recordHistory_)
                   lineGroup_.push_back(text.text_);

               return true;

           }

           // if we get here, the buffer has data in it but not the end of the line.

           text.text_.append(buffer_, &buffer_[bytesInBuffer_]);  // copy the data out of the buffer
                                                                  // into the target string.

           bytesInBuffer_ = 0;


        }



        // bytesInBuffer_ should be zero here or you broke the code...

        for(;;)
        {
            int bytesRead;

            if(bytesInBuffer_ != 0)
            {
                throw -1; // big time error
            }

            if(stringFile_)
            {   
               // we are reading from a stringstream not a real file

               if(stringEofEncountered_)
               {
                   good_ = false;
                   bytesRead = 0;
                   return !text.text_.empty();
               }
               else
               {
                   
                   stringFile_->read(buffer_, sizeof(buffer_));
                   
                   bytesRead = stringFile_->gcount();  // count of bytes read during last operation
                   
                   if(bytesRead == 0)
                   {
                      stringEofEncountered_ = false;
                   
                      if(!text.text_.empty())
                           if(recordHistory_)
                               lineGroup_.push_back(text.text_);
                      
                      return !text.text_.empty();
                   
                   }
               }

            }
            else
            {
#ifdef _MSC_VER
                            DWORD inputBytes=0;
                            
                            SetLastError(0);
                
                            if(!ReadFile(
                                          fh_,
                                          buffer_,
                                          sizeof(buffer_),
                                          &inputBytes,
                                          0
                                         )
                            )
                            {
                               // io error means return.
                
                                DWORD err = GetLastError();
                
                                CXXTLS_LOG("ReadFile error:  " << cxxtls::lastErrno());
                
                               good_ = false;
                               error_ = true;
                
                               if(text.text_.empty())
                                  return false;
                
                               if(recordHistory_)
                                   lineGroup_.push_back(text.text_);
                
                               return true; // the end of file be handled on the next call
                            }
                
                            if(inputBytes == 0)
                            {
                               good_ = false;
                
                               if(!text.text_.empty())
                                   if(recordHistory_)
                                       lineGroup_.push_back(text.text_);
                
                               return !text.text_.empty();
                            }
                
                            bytesRead = inputBytes;
                
                            if(buffer_[bytesRead-1] == '\r')
                            {
                               buffer_[bytesRead-1] = '\n';
                
                               if(isConsole_)
                               {
                                 cout << endl;
                               }
                            }
                            
#else
                            bytesRead = read(fh_, buffer_, sizeof(buffer_));
                
                            if(bytesRead <= 0)
                            {
                               good_ = false;
                
                               error_ = int(bytesRead) < 0;
                
                               if(!text.text_.empty())
                                    if(recordHistory_)
                                        lineGroup_.push_back(text.text_);
                
                               return !text.text_.empty();
                            }
#endif
            // if we get here, then is data in the buffer_.
           }

           bytesInBuffer_ = bytesRead;  

           char const *where = (char*)memchr(buffer_, '\n', bytesInBuffer_);

           if(where)
           {
               // there is a newline in the data in the buffer, lets return that
               // and remove those bytes from the buffer.

               text.text_.append(buffer_, where+1);   // use the data, include the \n in the data

               memcpy(buffer_, where+1, bytesInBuffer_ - (where-buffer_) - 1);  

               bytesInBuffer_ -= 1 + (where - buffer_);

               if(recordHistory_)
                   lineGroup_.push_back(text.text_);

               return true;

           }
           else
           {
              // data is in the buffer but there is no \n there -- so copy the data out and
              // reset the counters to 0

              text.text_.append(buffer_, &buffer_[bytesInBuffer_]);

              bytesInBuffer_ = 0;

           }


        }



    }  // readLine()

    void 
    ScriptLineStream::
    saveHistory()
    {
        string historyLine;

        list<string>::const_iterator f = lineGroup_.begin(),
                                     l = lineGroup_.end();
        while(f != l)
        {
            string cur = *f++;

            if(!cur.empty())
            {
               size_t end = cur.find_last_not_of("\r\n");

               if(end < cur.size())
                  cur.erase(end+1);
            }

            if(cur.find_first_not_of(" \n\r\t") >= cur.size())
              continue;


            historyLine += cur;

            if(f != l)
               historyLine += " ; ";
        }

        if (!historyLine.empty())
            history_.push_back(historyLine);

        lineGroup_.clear();

    }



} // namespace cxxtls
